
from gevent_fsm.fsm import State, transitions


class _Start(State):

    @transitions('Ready')
    def start(self, controller):

        controller.changeState(Ready)


Start = _Start()


class _Ready(State):

    def start(self, controller):
        print("server_fsm buffered_messages", len(
            controller.context.buffered_messages))
        if not controller.context.buffered_messages.empty():
            controller.context.queue.put(
                controller.context.buffered_messages.get())

    def onServiceInstance(self, controller, message_type, message):
        print(message)
        controller.context.start_or_update_monitor(message)


Ready = _Ready()


class _Waiting(State):

    pass


Waiting = _Waiting()





from deepdiff import DeepDiff, extract

import yaml
import re

def test_same():
    t1 = {1:1, 2:2, 3:3}
    t2 = t1.copy()
    assert DeepDiff(t1, t2) == {}
 
def test_different():
    t1 = {1:1, 2:2, 3:3}
    t2 = {1:2, 2:2, 3:3}
    assert DeepDiff(t1, t2) == {'values_changed': {'root[1]': {'new_value': 2, 'old_value': 1}}}


def test_diff_tree_add_item():

    t1 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
    ''')

    t2 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
        - name: R3
    ''')

    assert DeepDiff(t1, t2) ==  {'iterable_item_added': {"root['routers'][2]": {'name': 'R3'}}}


def test_diff_tree_remove_item():

    t1 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
    ''')

    t2 = yaml.safe_load('''
    routers:
        - name: R1
    ''')

    assert DeepDiff(t1, t2) ==  {'iterable_item_removed': {"root['routers'][1]": {'name': 'R2'}}}


def test_diff_tree_change_item():

    t1 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
    ''')

    t2 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R3
    ''')

    # Escape the pattern with a limited set of regexp special characters
    # To allow some of the regexps to be used.
    # This might now work out in the long term
    # We might want to use glob pattern matching or something simple like that
    _special_chars_map = {i: '\\' + chr(i) for i in b'()[]{}-|^$.&~# \t\n\r\v\f'}

    def escape(pattern):
        return pattern.translate(_special_chars_map)

    def make_matcher(pattern):
        pattern = escape(pattern)
        return f'^{pattern}.*'

    diff = DeepDiff(t1, t2)
    assert list(diff['values_changed'].keys())[0] == "root['routers'][1]['name']"
    assert re.escape("root['routers'][1]['name']") == r"root\['routers'\]\[1\]\['name'\]"
    assert re.match(re.escape("root['routers'][1]['name']"), "root['routers'][1]['name']")
    assert re.match(r"root\['routers'\]\[\d+\]\['name'\]", "root['routers'][1]['name']")
    assert re.match(r"root\['routers'\]\[.*\]\['name'\]", "root['routers'][1]['name']")
    assert re.match(r"^root\['routers'\]\[.*\]\['name'\].*", "root['routers'][1]['name']")
    assert re.match(r"^root\['routers'\].*", "root['routers'][1]['name']")
    assert re.match(make_matcher("root['routers']"), "root['routers'][1]['name']")
    assert re.match(make_matcher(r"root['routers'][\d+]"), "root['routers'][1]['name']")
    assert diff == {'values_changed': {"root['routers'][1]['name']": {'new_value': 'R3', 'old_value': 'R2'}}}


def test_extract_and_modify():

    t1 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
    ''')

    assert extract(t1, "root['routers'][0]") == {'name': 'R1'}
    extract(t1, "root['routers']")[0] = {'name': 'R3'}
    assert extract(t1, "root['routers'][0]") == {'name': 'R3'}

    # find parent and index of node


    # List case
    match = re.match(r"(.*)\[(\d+)\]$", "root['routers'][0]")
    assert match.groups()[0] == "root['routers']"
    assert match.groups()[1] == "0"
    parent = match.groups()[0]
    index = int(match.groups()[1])
    extract(t1, parent)[index] = {'name': 'R4'}

    assert extract(t1, "root['routers']") == [{'name': 'R4'}, {'name': 'R2'}]

    # Dictionary case
    match = re.match(r"(.*)\['(\S+)'\]$", "root['routers']")
    assert match.groups()[0] == "root"
    assert match.groups()[1] == "routers"
    parent = match.groups()[0]
    index = match.groups()[1]
    extract(t1, parent)[index] = [{'name': 'R5'}, {'name': 'R2'}]

    assert extract(t1, "root") == {'routers': [{'name': 'R5'}, {'name': 'R2'}]}





"""
Usage:
    desired-state [options] control [<control-id>]
    desired-state [options] monitor <current-state.yml> [<rules.yml>]
    desired-state [options] from <initial-state.yml> to <new-state.yml> [<rules.yml>]
    desired-state [options] update-desired-state <new-state.yml>
    desired-state [options] update-actual-state <new-state.yml>
    desired-state [options] validate <state.yml> <schema.yml>

Options:
    -h, --help              Show this page
    --debug                 Show debug logging
    --verbose               Show verbose logging
    --explain               Do not run the rules, only print the ones that would run.
    --ask-become-pass       Ask for the become password
    --project-src=<d>       Copy project files this directory [default: .]
    --inventory=<i>         Inventory to use
    --cwd=<c>               Change working directory on start
    --stream=<s>            Websocket channel to stream telemetry to
    --control-plane=<s>     Websocket channel to connect to the control plane
"""

from .stream import WebsocketChannel, NullChannel
from .messages import DesiredState, ActualState, Shutdown, now
from .util import ConsoleTraceLog, check_state
from .server import ZMQServerChannel
from .client import ZMQClientChannel
from .monitor import DesiredStateMonitor
from .control import DesiredStateControl
from .validate import get_errors, validate
from .collection import split_collection_name, has_rules, has_schema, load_rules, load_schema
from .types import get_meta
import gevent_fsm.conf
import gevent.exceptions
from getpass import getpass
from collections import defaultdict
from docopt import docopt
import yaml
import os
import sys
import logging
from uuid import uuid4
import gevent
from gevent import monkey
monkey.patch_all()


FORMAT = "[%(asctime)s] %(levelname)s [%(name)s.%(funcName)s:%(lineno)d] %(message)s"
logging.basicConfig(filename='/tmp/desired_state.log', level=logging.DEBUG, format=FORMAT)  # noqa
logging.debug('Logging started')
logging.debug('Loading runner')
logging.debug('Loaded runner')

logger = logging.getLogger('cli')


def main(args=None):
    '''
    Main function for the CLI.
    '''

    if args is None:
        args = sys.argv[1:]
    parsed_args = docopt(__doc__, args)
    if parsed_args['--debug']:
        logging.basicConfig(level=logging.DEBUG)
        gevent_fsm.conf.settings.instrumented = True
    elif parsed_args['--verbose']:
        gevent_fsm.conf.settings.instrumented = True
        logging.basicConfig(level=logging.INFO)
    else:
        logging.basicConfig(level=logging.WARNING)

    if parsed_args['--cwd']:
        os.chdir(parsed_args['--cwd'])

    if parsed_args['control']:
        return desired_state_control(parsed_args)
    elif parsed_args['monitor']:
        return desired_state_monitor(parsed_args)
    elif parsed_args['from'] and parsed_args['to']:
        return desired_state_from_to(parsed_args)
    elif parsed_args['update-desired-state']:
        return desired_state_update_desired_state(parsed_args)
    elif parsed_args['update-actual-state']:
        return desired_state_update_actual_state(parsed_args)
    elif parsed_args['validate']:
        return desired_state_validate(parsed_args)
    else:
        assert False, 'Update the docopt'


def inventory(parsed_args, state):
    '''
    Loads an inventory
    '''

    meta = get_meta(state)

    if meta.inventory and os.path.exists(meta.inventory):
        print('inventory:', meta.inventory)
        with open(meta.inventory) as f:
            return f.read()
    elif not parsed_args['--inventory']:
        print('inventory:', 'localhost only')
        return "all:\n  hosts:\n    localhost: ansible_connection=local\n"
    else:
        print('inventory:', parsed_args['--inventory'])
        with open(parsed_args['--inventory']) as f:
            return f.read()


def validate_state(state):
    '''
    Validates state using schema if it is found in the meta data of the state.
    '''

    meta = get_meta(state)

    if meta.schema:
        if os.path.exists(meta.schema):
            with open(meta.schema) as f:
                schema = yaml.safe_load(f.read())
        elif has_schema(*split_collection_name(meta.schema)):
            schema = load_schema(*split_collection_name(meta.schema))
        else:
            schema = {}
        validate(state, schema)


def parse_options(parsed_args):

    secrets = defaultdict(str)

    if parsed_args['--ask-become-pass'] and not secrets['become']:
        secrets['become'] = getpass()

    if parsed_args['--stream']:
        stream = WebsocketChannel(parsed_args['--stream'])
    else:
        stream = NullChannel()

    project_src = os.path.abspath(
        os.path.expanduser(parsed_args['--project-src']))

    return secrets, project_src, stream


def load_rules_from_args_or_meta(parsed_args, state):

    meta = get_meta(state)

    if parsed_args['<rules.yml>']:
        if os.path.exists(parsed_args['<rules.yml>']):
            with open(parsed_args['<rules.yml>']) as f:
                rules = yaml.safe_load(f.read())
        elif has_rules(*split_collection_name(parsed_args['<rules.yml>'])):
            rules = load_rules(
                *split_collection_name(parsed_args['<rules.yml>']))
        else:
            raise Exception('No rules file found')
    elif meta.rules:
        if os.path.exists(meta.rules):
            with open(meta.rules) as f:
                rules = yaml.safe_load(f.read())
        elif has_rules(*split_collection_name(meta.rules)):
            rules = load_rules(*split_collection_name(meta.rules))
        else:
            raise Exception('No rules file found')
    else:
        raise Exception('No rules file found')

    return rules


def desired_state_control(parsed_args):
    secrets, _, stream = parse_options(parsed_args)
    control_id = parsed_args['<control-id>'] or str(uuid4())

    if parsed_args['--control-plane']:
        control_plane = WebsocketChannel(parsed_args['--control-plane'])
    else:
        control_plane = NullChannel()

    threads = []

    if stream.thread:
        threads.append(stream.thread)

    tracer = ConsoleTraceLog()
    control = DesiredStateControl(
        tracer, 0, control_id, secrets, stream, control_plane)
    control_plane.outbox = control.queue
    threads.append(control.thread)

    server = ZMQServerChannel(control.queue, tracer)
    threads.append(server.zmq_thread)
    threads.append(server.controller_thread)
    control.controller.outboxes['output'] = server.queue
    gevent.joinall(threads)


def desired_state_monitor(parsed_args):
    '''
    Starts the state monitoring green thread.
    '''

    secrets, project_src, stream = parse_options(parsed_args)

    threads = []

    if stream.thread:
        threads.append(stream.thread)

    with open(parsed_args['<current-state.yml>']) as f:
        current_desired_state = yaml.safe_load(f.read())

    validate_state(current_desired_state)

    rules = load_rules_from_args_or_meta(parsed_args, current_desired_state)

    tracer = ConsoleTraceLog()
    worker = DesiredStateMonitor(tracer, 0, secrets, project_src, rules, current_desired_state, inventory(
        parsed_args, current_desired_state), stream)
    threads.append(worker.thread)
    server = ZMQServerChannel(worker.queue, tracer)
    threads.append(server.zmq_thread)
    threads.append(server.controller_thread)
    worker.controller.outboxes['output'] = server.queue
    gevent.joinall(threads)
    return 0


def desired_state_from_to(parsed_args):
    '''
    Calculates the differene in state from initial-state to new-state executes those changes and exits.
    '''

    secrets, project_src, stream = parse_options(parsed_args)

    threads = []

    if stream.thread:
        threads.append(stream.thread)

    with open(parsed_args['<initial-state.yml>']) as f:
        initial_desired_state = yaml.safe_load(f.read())

    validate_state(initial_desired_state)

    with open(parsed_args['<new-state.yml>']) as f:
        new_desired_state = f.read()

    validate_state(yaml.safe_load(new_desired_state))

    rules = load_rules_from_args_or_meta(parsed_args, initial_desired_state)

    tracer = ConsoleTraceLog()
    worker = DesiredStateMonitor(tracer, 0, secrets, project_src, rules, initial_desired_state, inventory(
        parsed_args, initial_desired_state), stream)
    threads.append(worker.thread)
    worker.queue.put(DesiredState(0, now(), 0, 0, new_desired_state))
    worker.queue.put(Shutdown())
    gevent.joinall([worker.thread])
    return 0


def desired_state_update_desired_state(parsed_args):
    '''
    Sends a new desired state to the monitor green thread.
    '''

    with open(parsed_args['<new-state.yml>']) as f:
        new_state = f.read()
        check_state(new_state)

    validate_state(yaml.safe_load(new_state))

    client = ZMQClientChannel()
    client.send(DesiredState(0, now(), 0, 0, new_state))
    return 0


def desired_state_update_actual_state(parsed_args):
    '''
    Sends a new actual state to the monitor green thread.
    '''

    with open(parsed_args['<new-state.yml>']) as f:
        new_state = f.read()
        check_state(new_state)

    validate_state(yaml.safe_load(new_state))

    client = ZMQClientChannel()
    client.send(ActualState(0, 0, new_state))
    return 0


def desired_state_validate(parsed_args):
    '''
    Validates a state using the schema and prints a list of errors in the state.
    '''

    with open(parsed_args['<state.yml>']) as f:
        state = yaml.safe_load(f.read())

    with open(parsed_args['<schema.yml>']) as f:
        schema = yaml.safe_load(f.read())

    for error in get_errors(state, schema):
        print(error)
    else:
        return 0
    return 1



if __name__ == "__main__":
    main()





from collections import namedtuple
import yaml
import json
import itertools
import datetime


def sequence():
    '''
    Sequence numbers are associated with a stream.  They are used to determine
    the order of the messages sent in a stream.
    '''
    return itertools.count(1)


def now():
    '''
    Returns a string formatted timestamp for now in UTC time.
    '''
    return datetime.datetime.now(datetime.timezone.utc).isoformat()


def serialize(message):
    '''
    Serializes a message to YAML format.
    '''
    return [message.__class__.__name__.encode(), yaml.dump(dict(message._asdict())).encode()]


def json_serialize(message):
    '''
    Serializes a message to JSON format.
    '''
    return json.dumps([message.__class__.__name__, dict(message._asdict())]).encode()


def json_deserialize(message):
    '''
    Deserializes a message from JSON format.
    '''
    data = json.loads(message)
    if isinstance(data, list):
        msg_type = data[0]
        msg_data = data[1]
        if msg_type in msg_types:
            try:
                return msg_types[msg_type](**msg_data)
            except BaseException as e:
                print(e)
                raise
    return None


Hello = namedtuple('Hello', ['seq_num', 'timestamp'])
FSMState = namedtuple('FSMState', ['seq_num', 'timestamp', 'state'])
Diff = namedtuple('Diff', ['seq_num', 'timestamp', 'diff'])
ValidationResult = namedtuple('ValidationResult', ['seq_num', 'timestamp', 'host', 'result'])
ValidationTask = namedtuple(
    'ValidationTask', ['seq_num', 'timestamp', 'host', 'task_action', 'result'])
Stdout = namedtuple('Stdout', ['seq_num', 'timestamp', 'stdout'])


DesiredState = namedtuple('DesiredState', ['seq_num', 'timestamp', 'id', 'client_id', 'desired_state'])
ActualState = namedtuple('ActualState', ['id', 'client_id', 'actual_state'])
Poll = namedtuple('Poll', [])
Complete = namedtuple('Complete', [])
Difference = namedtuple('Difference', [])
NoDifference = namedtuple('NoDifference', [])
Success = namedtuple('Success', [])
Failure = namedtuple('Failure', [])

Inventory = namedtuple('Inventory', ['seq_num', 'timestamp', 'inventory'])
Rules = namedtuple('Rules', ['seq_num', 'timestamp', 'rules'])

Control = namedtuple('Control', ['seq_num', 'timestamp', 'id'])
System = namedtuple('System', ['id', 'control_id'])
Monitor = namedtuple('Monitor', ['id', 'system_id', 'control_id'])

DesiredSystemState = namedtuple(
    'DesiredSystemState', ['id', 'client_id', 'desired_state'])

Shutdown = namedtuple('Shutdown', [])

ServiceInstance = namedtuple('ServiceInstance', ['id',
                                                 'service_id',
                                                 'created_at',
                                                 'deleted_at',
                                                 'name',
                                                 'config',
                                                 'inventory',
                                                 'inventory_id',
                                                 'collection',
                                                 'service_name',
                                                 'schema_name',
                                                 'rules_name',
                                                 'status'])

msg_types = {x.__name__: x for x in [
    DesiredState, ActualState, Hello, Control, ServiceInstance]}




import pytest
from desired_state.validate import validate, get_errors, ValidationError


def test_validate_empty():
    validate({}, {})


def test_get_errors_empty():
    errors = get_errors({}, {})
    assert len(list(errors)) == 0


def test_validate_simple():
    schema = {
        "type": "object",
        "properties": {
            "price": {"type": "number"},
            "name": {"type": "string"},
        },
    }
    validate({"name" : "Eggs", "price" : 34.99}, schema)
    with pytest.raises(ValidationError):
        assert validate({"name" : "Eggs", "price" : "Invalid"}, schema)


def test_get_errors_simple():
    schema = {
        "type": "object",
        "properties": {
            "price": {"type": "number"},
            "name": {"type": "string"},
        },
    }
    errors = get_errors({"name" : "Eggs", "price" : "Invalid"}, schema)
    assert len(list(errors)) == 1





import re
from .util import make_matcher, build_rule_selector
from enum import Enum
from deepdiff import extract
from collections import OrderedDict


class Action(Enum):

    CREATE = 'CREATE'
    UPDATE = 'UPDATE'
    RETRIEVE = 'RETRIEVE'
    DELETE = 'DELETE'
    RENAME = 'RENAME'
    VALIDATE = 'VALIDATE'


ACTION_RULES = {Action.CREATE: 'create',
                Action.UPDATE: 'update',
                Action.RETRIEVE: 'retrieve',
                Action.DELETE: 'delete',
                Action.RENAME: 'rename',
                Action.VALIDATE: 'validate'}


def select_rules_recursive(diff, rules, current_desired_state, new_desired_state):

    matching_rules = []
    matchers = [(make_matcher(build_rule_selector(
        rule['rule_selector'])), rule) for rule in rules]

    for key, value in diff.get('values_changed', {}).items():
        for (matcher, rule) in matchers:
            match = re.match(matcher, key)
            if match:
                matching_rules.append(('values_changed', rule, match, value))

    for item in diff.get('dictionary_item_added', []):
        for (matcher, rule) in matchers:
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('dictionary_item_added', rule, match, None))
            new_subtree = extract(new_desired_state, item)
            select_rules_recursive_helper(
                diff, matchers, matching_rules, item, new_subtree)

    for item in diff.get('dictionary_item_removed', []):
        for (matcher, rule) in matchers:
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('dictionary_item_removed', rule, match, None))

    for item in diff.get('iterable_item_added', []):
        print(item)
        for (matcher, rule) in matchers:
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('iterable_item_added', rule, match, None))

    for item in diff.get('iterable_item_removed', []):
        print(item)
        for (matcher, rule) in matchers:
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('iterable_item_removed', rule, match, None))

    for key, value in diff.get('type_changes', {}).items():
        # Handles case in YAML where an empty list defaults to None type
        if value.get('old_type') == type(None) and value.get('new_type') == list:
            # Add a single new element to the key
            # TODO: this should probably loop over all the new elements in the list not just one
            key += '[0]'
        elif value.get('old_type') == list and value.get('new_type') == type(None):
            # Add a single new element to the key
            key += '[0]'
        for (matcher, rule) in matchers:
            match = re.match(matcher, key)
            if match:
                matching_rules.append(('type_changes', rule, match, None))

        # Handles case in YAML where an empty dict defaults to None type on the old state
        if value.get('old_type') == type(None) and value.get('new_type') == dict:
            # Try the matcher against all the keys in the dict
            for dict_key in value.get('new_value').keys():
                new_key = f"{key}['{dict_key}']"
                for (matcher, rule) in matchers:
                    match = re.match(matcher, new_key)
                    if match:
                        matching_rules.append(
                            ('type_changes', rule, match, None))
            select_rules_recursive_helper(
                diff, matchers, matching_rules, key, value.get('new_value'))

        # Handles case in YAML where an empty dict defaults to None type on the new state
        if value.get('old_type') == dict and value.get('new_type') == type(None):
            # Try the matcher against all the keys in the dict
            for dict_key in value.get('old_value').keys():
                old_key = f"{key}['{dict_key}']"
                for (matcher, rule) in matchers:
                    match = re.match(matcher, old_key)
                    if match:
                        matching_rules.append(
                            ('type_changes', rule, match, None))
            select_rules_recursive_helper(
                diff, matchers, matching_rules, key, value.get('old_value'))

    return matching_rules


def select_rules_recursive_helper(diff, matchers, matching_rules, path, value):

    for (matcher, rule) in matchers:
        match = re.match(matcher, path)
        if match:
            matching_rules.append(('subtree', rule, match, value))

    if type(value) is list:
        for i, item in enumerate(value):
            select_rules_recursive_helper(
                diff, matchers, matching_rules, f"{path}[{i}]", item)

    if type(value) is dict:
        for k, v in value.items():
            select_rules_recursive_helper(
                diff, matchers, matching_rules, f"{path}['{k}']", v)


def select_rules(diff, rules):
    matching_rules = []
    for rule in rules:
        matcher = make_matcher(build_rule_selector(rule['rule_selector']))
        for key, value in diff.get('values_changed', {}).items():
            match = re.match(matcher, key)
            if match:
                matching_rules.append(('values_changed', rule, match, value))
        for item in diff.get('dictionary_item_added', []):
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('dictionary_item_added', rule, match, None))
        for item in diff.get('dictionary_item_removed', []):
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('dictionary_item_removed', rule, match, None))
        for item in diff.get('iterable_item_added', []):
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('iterable_item_added', rule, match, None))
        for item in diff.get('iterable_item_removed', []):
            match = re.match(matcher, item)
            if match:
                matching_rules.append(
                    ('iterable_item_removed', rule, match, None))
        for key, value in diff.get('type_changes', {}).items():
            # Handles case in YAML where an empty list defaults to None type
            if value.get('old_type') == type(None) and value.get('new_type') == list:
                # Add a single new element to the key
                # TODO: this should probably loop over all the new elements in the list not just one
                key += '[0]'
            elif value.get('old_type') == list and value.get('new_type') == type(None):
                # Add a single new element to the key
                key += '[0]'
            match = re.match(matcher, key)
            if match:
                matching_rules.append(('type_changes', rule, match, None))

            # Handles case in YAML where an empty dict defaults to None type
            if value.get('old_type') == type(None) and value.get('new_type') == dict:
                # Try the matcher against all the keys in the dict
                for dict_key in value.get('new_value').keys():
                    new_key = f"{key}['{dict_key}']"
                    match = re.match(matcher, new_key)
                    if match:
                        matching_rules.append(
                            ('type_changes', rule, match, None))

    return matching_rules


def deduplicate_rules(matching_rules):
    # Deduplicate the rules since some rules may match more than once when using recursive rule selection

    dedup_matching_rules = OrderedDict()

    for matching_rule in matching_rules:
        _, _, match, _ = matching_rule
        changed_subtree_path = match.groups()[0]
        if changed_subtree_path not in dedup_matching_rules:
            dedup_matching_rules[changed_subtree_path] = matching_rule

    dedup_matching_rules = list(dedup_matching_rules.values())

    return dedup_matching_rules


def get_rule_action_subtree(matching_rule, current_desired_state, new_desired_state):

    change_type, rule, match, value = matching_rule
    print('change_type', change_type)
    print('rule', rule)
    print('match', match)
    print('value', value)
    changed_subtree_path = match.groups()[0]
    print('changed_subtree_path', changed_subtree_path)
    try:
        new_subtree = extract(new_desired_state, changed_subtree_path)
        new_subtree_missing = False
    except (KeyError, IndexError, TypeError):
        new_subtree_missing = True
    try:
        old_subtree = extract(current_desired_state, changed_subtree_path)
        old_subtree_missing = False
    except (KeyError, IndexError, TypeError):
        old_subtree_missing = True
    print('new_subtree_missing', new_subtree_missing)
    print('old_subtree_missing', old_subtree_missing)

    if change_type == 'iterable_item_added':
        action = Action.CREATE
        subtree = new_subtree
    elif change_type == 'iterable_item_removed':
        action = Action.DELETE
        subtree = old_subtree
    elif new_subtree_missing is False and old_subtree_missing is False:
        action = Action.UPDATE
        subtree = new_subtree
    elif new_subtree_missing and old_subtree_missing is False:
        action = Action.DELETE
        subtree = old_subtree
    elif old_subtree_missing and new_subtree_missing is False:
        action = Action.CREATE
        subtree = new_subtree
    else:
        assert False, "Logic bug"
    print('action', action)

    return action, subtree



from gevent_fsm.fsm import State, transitions

import yaml
from deepdiff import DeepDiff
from pprint import pprint
from .diff import desired_state_diff, desired_state_discovery, desired_state_validation, convert_diff
from .messages import FSMState, DesiredState, Diff, now


class _Validate1(State):

    @transitions('Diff2')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Validate1'))

        # Trivial discovery
        # Assume the state is the same as new desired state

        monitor = controller.context

        monitor.operational_actual_state = desired_state_validation(monitor,
                                                                    monitor.secrets,
                                                                    monitor.project_src,
                                                                    monitor.new_desired_state,
                                                                    monitor.ran_rules,
                                                                    monitor.inventory,
                                                                    False)
        controller.changeState(Waiting)


Validate1 = _Validate1()


class _Discover1(State):

    @transitions('Diff2')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Discover1'))

        # Trivial discovery
        # Assume the state is the same as new desired state

        monitor = controller.context

        monitor.discovered_actual_state = desired_state_discovery(monitor,
                                                                  monitor.secrets,
                                                                  monitor.project_src,
                                                                  monitor.current_desired_state,
                                                                  monitor.new_desired_state,
                                                                  monitor.ran_rules,
                                                                  monitor.inventory,
                                                                  False)

        controller.changeState(Diff2)


Discover1 = _Discover1()


class _Help(State):

    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Help'))


Help = _Help()


class _Reconcile1(State):

    @transitions('Discover1')
    @transitions('Retry')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Reconcile1'))

        monitor = controller.context

        monitor.ran_rules = desired_state_diff(monitor,
                                               monitor.secrets,
                                               monitor.project_src,
                                               monitor.current_desired_state,
                                               monitor.new_desired_state,
                                               monitor.rules,
                                               monitor.inventory,
                                               False)

        if True:
            controller.changeState(Discover1)
        else:
            controller.changeState(Retry)


Reconcile1 = _Reconcile1()


class _Reconcile2(State):

    @transitions('Discover1')
    @transitions('Retry')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Reconcile2'))

        monitor = controller.context

        result = desired_state_diff(monitor,
                                    monitor.secrets,
                                    monitor.project_src,
                                    monitor.discovered_actual_state,
                                    monitor.new_desired_state,
                                    monitor.rules,
                                    monitor.inventory,
                                    False)

        if result:
            controller.changeState(Discover1)
        else:
            controller.changeState(Retry)


Reconcile2 = _Reconcile2()


class _Reconcile3(State):

    @transitions('Discover1')
    @transitions('Retry')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Reconcile3'))

        monitor = controller.context

        result = desired_state_diff(monitor,
                                    monitor.secrets,
                                    monitor.project_src,
                                    monitor.discovered_actual_state,
                                    monitor.current_desired_state,
                                    monitor.rules,
                                    monitor.inventory,
                                    False)

        if result:
            controller.changeState(Discover2)
        else:
            controller.changeState(Retry)


Reconcile3 = _Reconcile3()


class _Waiting(State):

    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Waiting'))
        print("reconciliation_fsm buffered_messages",
              len(controller.context.buffered_messages))
        if not controller.context.buffered_messages.empty():
            controller.context.queue.put(
                controller.context.buffered_messages.get())

    @transitions('Diff1')
    def onDesiredState(self, controller, message_type, message):
        print('Waiting.onDesiredState')
        controller.context.new_desired_state = yaml.safe_load(
            message.desired_state)
        controller.context.stream.put_message(
            DesiredState(0, now(), 0, 0, controller.context.new_desired_state))
        controller.changeState(Diff1)

    @transitions('Diff1')
    def onActualState(self, controller, message_type, message):
        print('Waiting.onActualState')
        controller.context.discovered_actual_state = yaml.safe_load(
            message.actual_state)
        controller.changeState(Diff3)

    @transitions('Discover2')
    def onPoll(self, controller, message_type, message):
        controller.changeState(Discover2)

    def onShutdown(self, controller, message_type, message):
        controller.context.thread.kill()


Waiting = _Waiting()


class _Diff1(State):

    @transitions('Reconcile1')
    @transitions('Waiting')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Diff1'))
        controller.context.diff = DeepDiff(
            controller.context.current_desired_state, controller.context.new_desired_state, ignore_order=True)
        pprint(controller.context.diff)
        controller.context.stream.put_message(
            Diff(0, now(), convert_diff(controller.context.diff)))

        if controller.context.diff:
            controller.changeState(Reconcile1)
        else:
            controller.changeState(Waiting)


Diff1 = _Diff1()


class _Revert(State):

    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Revert'))

    @transitions('Help')
    def failure(self, controller, message_type, message):

        controller.changeState(Help)

    @transitions('Discover1')
    def success(self, controller, message_type, message):

        controller.changeState(Discover1)


Revert = _Revert()


class _Diff3(State):

    @transitions('Reconcile3')
    @transitions('Waiting')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Diff3'))
        controller.context.diff = DeepDiff(
            controller.context.discovered_actual_state, controller.context.current_desired_state, ignore_order=True)
        print(controller.context.diff)

        if controller.context.diff:
            controller.changeState(Reconcile3)
        else:
            controller.changeState(Waiting)


Diff3 = _Diff3()


class _Discover2(State):

    @transitions('Diff3')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Discover2'))

        # Trivial discovery
        # Assume the state is the same as current desired state
        controller.context.discovered_actual_state = controller.context.current_desired_state
        controller.changeState(Diff3)


Discover2 = _Discover2()


class _Start(State):

    @transitions('Waiting')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Start'))
        controller.changeState(Waiting)


Start = _Start()


class _Diff2(State):

    @transitions('Reconcile2')
    @transitions('Validate1')
    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Diff2'))
        controller.context.diff = DeepDiff(
            controller.context.new_desired_state, controller.context.discovered_actual_state, ignore_order=True)
        print(controller.context.diff)
        controller.context.stream.put_message(
            Diff(0, now(), convert_diff(controller.context.diff)))

        if controller.context.diff:
            controller.changeState(Reconcile2)
        else:
            controller.context.current_desired_state = controller.context.new_desired_state
            controller.changeState(Validate1)


Diff2 = _Diff2()


class _Retry(State):

    def start(self, controller):
        controller.context.stream.put_message(FSMState(0, now(), 'Retry'))

    @transitions('Revert')
    def failure(self, controller, message_type, message):

        controller.changeState(Revert)

    @transitions('Discover1')
    def success(self, controller, message_type, message):

        controller.changeState(Discover1)


Retry = _Retry()





import jsonschema


def get_errors(state, schema):
    validator = jsonschema.Draft7Validator(schema)
    for error in validator.iter_errors(state):
        yield error


def validate(state, schema):
    validator = jsonschema.Draft7Validator(schema)
    validator.validate(state)


ValidationError = jsonschema.ValidationError





import gevent
import yaml
from gevent.queue import Queue
from gevent_fsm.fsm import FSMController, Channel

from . import reconciliation_fsm

from .messages import Inventory, Rules, DesiredState, now


def convert_inventory(inventory):
    '''
        Converts an inventory from a YAML string to a data structure.
    '''
    inventory = yaml.safe_load(inventory)
    return inventory


class DesiredStateMonitor(object):

    '''
    A monitor is a green thread FSM that monitors the lifecycle of one
    desired state configuration defined by one schema.   The FSM is defined by
    the reconciliation_fsm.py file.  The monitor receives state changes from
    its `queue` and sends status messages out of its `stream`.
    '''

    def __init__(self, tracer, fsm_id, secrets, project_src, rules, current_desired_state, inventory, stream):
        self.secrets = secrets
        self.project_src = project_src
        self.rules = rules
        self.ran_rules = []
        self.new_desired_state = None
        self.current_desired_state = current_desired_state
        self.discovered_actual_state = None
        self.operational_actual_state = None
        self.inventory = inventory
        self.tracer = tracer
        self.stream = stream
        self.buffered_messages = Queue()
        self.controller = FSMController(
            self, "reconciliation_fsm", fsm_id, reconciliation_fsm.Start, self.tracer, self.tracer)
        self.controller.outboxes['default'] = Channel(
            self.controller, self.controller, self.tracer, self.buffered_messages)
        self.queue = self.controller.inboxes['default']
        self.stream.put_message(Inventory(0, now(), convert_inventory(inventory)))
        self.stream.put_message(Rules(0, now(), rules))
        self.stream.put_message(DesiredState(0, now(), 0, 0, current_desired_state))
        self.thread = gevent.spawn(self.controller.receive_messages)




from collections import namedtuple

Meta = namedtuple('Meta', ['rules', 'schema', 'inventory'])


def get_meta(state):
    '''
    Gets the meta data from a state file.
    '''
    if not state:
        return Meta(None, None, None)
    elif 'meta' in state:
        return Meta(state['meta'].get('rules', None),
                    state['meta'].get('schema', None),
                    state['meta'].get('inventory', None))
    else:
        return Meta(None, None, None)





import zmq
from .messages import serialize


class ZMQClientChannel(object):

    def __init__(self):
        self.context = zmq.Context.instance()
        self.socket = self.context.socket(zmq.DEALER)
        self.socket.connect('tcp://127.0.0.1:5555')

    def send(self, task):
        self.socket.send_multipart(serialize(task))
        msg = self.socket.recv_multipart()
        print(msg)

    def receive(self):
        return self.socket.recv_multipart()





import gevent
import yaml
from gevent.queue import Queue
from gevent_fsm.fsm import FSMController, Channel

from . import control_fsm

from .messages import Control, DesiredState, now
from .monitor import DesiredStateMonitor
from .collection import load_rules, load_schema, split_collection_name
from .validate import validate


class DesiredStateControl(object):

    '''
    A control is a green thread FSM that controls multiple monitors (monitor.py) in one or more
    systems.   Each system is managed by one or more monitors and is defined by a set of desired
    state configurations.  A set of desired state configurations that work together defines a system.
    For instance a set of linux, networking, and application desired state configurations would
    define a system in this terminology.  The control would bring up the linux, networking, and
    application configurations in the correct order to bring the system from zero to a completely
    working application.
    '''

    def __init__(self, tracer, fsm_id, control_id, secrets, stream, control_plane):
        self.service_instances = dict()
        self.workers = dict()
        self.control_id = control_id
        self.secrets = secrets
        self.tracer = tracer
        self.stream = stream
        self.control_plane = control_plane
        self.buffered_messages = Queue()
        self.controller = FSMController(
            self, "control_fsm", fsm_id, control_fsm.Start, self.tracer, self.tracer)
        self.controller.outboxes['default'] = Channel(
            self.controller, self.controller, self.tracer, self.buffered_messages)
        self.queue = self.controller.inboxes['default']
        self.control_plane.put_message(Control(0, now(), self.control_id))
        self.stream.put_message(Control(0, now(), self.control_id))
        self.control_plane.reconnect_callback = self.reconnect
        self.stream.reconnect_callback = self.reconnect
        self.thread = gevent.spawn(self.controller.receive_messages)

    def reconnect(self, stream):
        stream.put_message(Control(0, now(), self.control_id))

    def start_monitor(self, service_instance):
        # Get schema
        schema = load_schema(
            *split_collection_name(service_instance.schema_name))
        print(schema)
        # Get rules
        rules = load_rules(*split_collection_name(service_instance.rules_name))
        print(rules)
        # Get inventory
        inventory = service_instance.inventory
        print(inventory)
        validate(yaml.safe_load(service_instance.config), schema)
        project_src = '.'
        worker = DesiredStateMonitor(self.tracer,
                                     0,
                                     self.secrets,
                                     project_src,
                                     rules,
                                     yaml.safe_load(service_instance.config),
                                     inventory,
                                     self.stream)
        self.workers[service_instance.id] = worker

        self.service_instances[service_instance.id] = service_instance

    def update_monitor(self, service_instance):
        schema = load_schema(
            *split_collection_name(service_instance.schema_name))
        validate(yaml.safe_load(service_instance.config), schema)
        self.workers[service_instance.id].queue.put(
            DesiredState(0, now(), 0, 0, service_instance.config))

    def start_or_update_monitor(self, service_instance):
        if service_instance.id in self.workers:
            self.update_monitor(service_instance)
        else:
            self.start_monitor(service_instance)



import zmq.green as zmq
from gevent.queue import Queue
from gevent_fsm.fsm import FSMController, Channel, NullChannel
from .messages import msg_types
import gevent
import yaml
from . import server_fsm


class ZMQServerChannel(object):

    def __init__(self, outbox, tracer):
        self.tracer = tracer
        self.buffered_messages = Queue()
        self.outbox = outbox
        self.context = zmq.Context.instance()
        self.socket = self.context.socket(zmq.ROUTER)
        self.socket.bind('tcp://127.0.0.1:5555')
        self.controller = FSMController(self,
                                        "server_fsm",
                                        1,
                                        server_fsm.Start,
                                        self.tracer,
                                        self.tracer)
        self.controller.outboxes['default'] = Channel(self.controller,
                                                      self.controller,
                                                      self.tracer,
                                                      self.buffered_messages)
        self.controller.outboxes['output'] = NullChannel(self.controller, self)
        self.queue = self.controller.inboxes['default']

        self.zmq_thread = gevent.spawn(self.receive_messages)
        self.controller_thread = gevent.spawn(self.controller.receive_messages)

    def receive_messages(self):
        while True:
            message = self.socket.recv_multipart()
            id = message.pop(0)
            msg_type = message.pop(0).decode()
            msg_data = yaml.safe_load(message.pop(0).decode())
            print(id, msg_type, msg_data)
            msg_data['client_id'] = id
            print(msg_types[msg_type](**msg_data))
            self.queue.put(msg_types[msg_type](**msg_data))
            self.socket.send_multipart([id, b'Submitted'])



import os
import yaml

HERE = os.path.abspath(os.path.dirname(__file__))


def load_rule(name):
    with open(os.path.join(HERE, 'rules', f'{name}.yml')) as f:
        return yaml.safe_load(f.read())


def load_state(name, version):
    with open(os.path.join(HERE, 'states', name, f'{version}.yml')) as f:
        return yaml.safe_load(f.read())


def load_system(name, version):
    with open(os.path.join(HERE, 'systems', name, f'{version}.yml')) as f:
        return yaml.safe_load(f.read())





from desired_state.messages import json_serialize, json_deserialize, Hello, Control
from desired_state.messages import sequence, now

import datetime


def test_hello():
    t = now()
    s = json_serialize(Hello(0, t))
    o = json_deserialize(s)
    assert o == Hello(0, t)


def test_control():
    t = now()
    s = json_serialize(Control(0, t, 5))
    o = json_deserialize(s)
    assert o == Control(0, t, 5)


def test_sequence():
    s = sequence()
    assert next(s) == 1
    assert next(s) == 2


def test_now():
    t = now()
    assert isinstance(t, str)
    assert datetime.datetime.fromisoformat(t)



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.md') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [ ]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Ben Thomasson",
    author_email='bthomass@redhat.com',
    python_requires='>=3.5',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: Apache Software License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
    ],
    description="A desired state system for ansible",
    install_requires=requirements,
    license="Apache Software License 2.0",
    long_description=readme + '\n\n' + history,
    include_package_data=True,
    keywords='desired_state',
    name='desired_state',
    packages=find_packages(include=['desired_state', 'desired_state.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/benthomasson/desired_state',
    version='0.1.0',
    zip_safe=False,
    entry_points={
        'console_scripts': [
            'desired-state = desired_state.cli:main',
        ],
    }
)





from desired_state.rule import build_rule_selector



def test_selector_normal():
    assert r"root['key'][\d+]" == build_rule_selector('root.key.index')


def test_selector_no_root():
    assert r"['key'][\d+]" == build_rule_selector('key.index')


def test_selector_many_keys():
    assert r"root['key']['foo']['bar']" == build_rule_selector('root.key.foo.bar')

def test_selector_many_indexes():
    assert r"root['key'][\d+][\d+]" == build_rule_selector('root.key.index.index')

def test_selector_name():
    assert r"['name']" == build_rule_selector('name')




import websocket
import gevent
import ssl
from pprint import pprint

from .messages import Hello, json_serialize, json_deserialize, now, sequence


class WebsocketChannel(object):

    def __init__(self, address):
        self.sequence = sequence()
        self.address = address
        self.thread = None
        self.reconnect_callback = None
        self.start_socket_thread()
        self.startup_messages = []
        self.opened = False
        self.outbox = None
        self.put_message(Hello(0, now()))

    def start_socket_thread(self):
        print('start_socket_thread')
        self.socket = websocket.WebSocketApp(self.address,
                                             on_message=self.on_message,
                                             on_error=self.on_error,
                                             on_close=self.on_close,
                                             on_open=self.on_open)
        self.thread = gevent.spawn(self.socket.run_forever, sslopt={
                                   "cert_reqs": ssl.CERT_NONE})

    def put_message(self, message):
        message = message._replace(seq_num=next(self.sequence))
        self.put(json_serialize(message))

    def put(self, message):
        pprint(message)
        if not self.opened:
            self.startup_messages.append(message)
        else:
            self.socket.send(message)

    def on_open(self, ws=None):
        print('on_open')
        self.opened = True
        for message in self.startup_messages:
            self.put(message)
        self.startup_messages = []

    def on_message(self, message):
        print('on_message')
        if self.outbox:
            msg = json_deserialize(message)
            if not msg:
                print("Unknown message:", message)
            self.outbox.put(msg)

    def on_close(self, ws=None):
        print('on_close')
        self.opened = False
        old_thread = self.thread
        self.start_socket_thread()
        self.put_message(Hello(0, now()))
        if self.reconnect_callback:
            self.reconnect_callback(self)
        old_thread.kill()

    def on_error(self, ws=None, error=None):
        print('WebsocketChannel on_error', error)
        self.on_close(ws)
        gevent.sleep(1)


class NullChannel(object):

    thread = None
    outbox = None

    def put(self, message):
        pass

    def put_message(self, message):
        pass




from itertools import count
import yaml
import os

# Escape the pattern with a limited set of regexp special characters
# To allow some of the regexps to be used.
# This might now work out in the long term
# We might want to use glob pattern matching or something simple like that
_special_chars_map = {i: '\\' + chr(i) for i in b'()[]{}-|^$.&~# \t\n\r\v\f'}


def escape(pattern):
    return pattern.translate(_special_chars_map)


def make_matcher(pattern):
    pattern = escape(pattern)
    return f'^({pattern}).*$'


def ensure_directory(d):
    if not os.path.exists(d):
        os.mkdir(d)


def check_state(state):
    yaml.safe_load(state)


class ConsoleTraceLog(object):

    def __init__(self):
        self.counter = count(start=1, step=1)

    def trace_order_seq(self):
        return next(self.counter)

    def send_trace_message(self, message):
        print(message)


def build_rule_selector(dotted_selector):
    '''
    The dotted selector should have the form of:
        root
        root.key
        root.key[.key]...
        root.key.index

    Where [.key]... means a repeating element
    index is a special key that means a list index
    root is a special key that only matches the root
    of the state.

    This function will return a regular expression
    that can be fed into deepdiff.extract.
    '''

    selector = []

    parts = dotted_selector.split('.')

    if parts[0] == "root":
        selector.append('root')
        start = 1
    elif parts[0] == "node":
        selector.append('node')
        start = 1
    else:
        start = 0

    for part in parts[start:]:
        if part == "index":
            selector.append(r'[\d+]')
        else:
            selector.append(f"['{part}']")

    return "".join(selector)


build_inventory_selector = build_rule_selector




import os
import yaml
import tempfile
import shutil
import json
import re
import glob
import ansible_runner
from pprint import pprint
from deepdiff import DeepDiff, extract

from .rule import select_rules_recursive, Action, ACTION_RULES, get_rule_action_subtree, deduplicate_rules
from .util import ensure_directory, build_inventory_selector
from .messages import ValidationResult, ValidationTask, now, Stdout
from .collection import split_collection_name, has_tasks, load_tasks


def convert_diff(diff):
    '''
    Converts the DeepDiff structure into a YAML serializable data structure.
    '''

    print(diff)
    if 'dictionary_item_added' in diff:
        diff['dictionary_item_added'] = [
            str(x) for x in diff['dictionary_item_added']]
    if 'dictionary_item_removed' in diff:
        diff['dictionary_item_removed'] = [
            str(x) for x in diff['dictionary_item_removed']]
    if 'type_changes' in diff:
        diff['type_changes'] = [str(x) for x in diff['type_changes']]
    diff = dict(diff)
    print(yaml.safe_dump(diff))
    return diff


class PlaybookRunner:

    '''
    PlaybookRunner is responsible for setting up and running ansible-runner
    '''

    def __init__(self, message_processor, new_desired_state, state_diff, destructured_vars_list, playbook, secrets, project_src, inventory):
        print('PlaybookRunner')
        self.message_processor = message_processor
        self.inventory = inventory
        self.secrets = secrets
        self.project_src = project_src
        self.new_desired_state = new_desired_state
        self.state_diff = convert_diff(state_diff)
        self.destructured_vars_list = destructured_vars_list
        self.playbook = playbook
        self.runner_thread = None
        self.shutdown_requested = False
        self.shutdown = False

    def run(self):

        self.build_project_directory()
        self.copy_files()
        self.build_project_subdirectories()
        self.write_settings()
        self.write_cmdline()
        self.write_passwords()
        self.write_state_vars()
        self.write_diff_vars()
        self.write_destructred_vars()
        self.write_playbook()
        self.write_inventory()
        self.start_ansible_playbook()
        return self.read_result()

    def build_project_directory(self):
        self.temp_dir = tempfile.mkdtemp(prefix="desired_state_playbook")
        print(self.temp_dir)
        ensure_directory(os.path.join(self.temp_dir, 'env'))
        ensure_directory(os.path.join(self.temp_dir, 'project'))

    def build_project_subdirectories(self):
        ensure_directory(os.path.join(self.temp_dir, 'project', 'roles'))

    def copy_files(self):
        src = os.path.abspath(self.project_src)
        dest = os.path.join(self.temp_dir, 'project')
        src_files = os.listdir(src)
        for file_name in src_files:
            full_file_name = os.path.join(src, file_name)
            if (os.path.isfile(full_file_name)):
                shutil.copy(full_file_name, dest)
            if (os.path.isdir(full_file_name)):
                shutil.copytree(full_file_name, os.path.join(dest, file_name))

    def write_settings(self):
        with open(os.path.join(self.temp_dir, 'env', 'settings'), 'w') as f:
            f.write(json.dumps(dict(idle_timeout=0,
                                    job_timeout=0)))

    def write_cmdline(self):
        with open(os.path.join(self.temp_dir, 'env', 'cmdline'), 'w') as f:
            f.write("--ask-become-pass -v")

    def write_passwords(self):
        with open(os.path.join(self.temp_dir, 'env', 'passwords'), 'w') as f:
            f.write(
                """---\n"SUDO password:": "{0}"\nBECOME password: "{0}"\n...""".format(self.secrets['become']))

    def write_playbook(self):
        self.playbook_file = (os.path.join(
            self.temp_dir, 'project', 'playbook.yml'))
        playbook = self.playbook
        with open(self.playbook_file, 'w') as f:
            f.write(yaml.safe_dump(playbook, default_flow_style=False))

    def write_state_vars(self):
        state_vars_file = os.path.join(
            self.temp_dir, 'project', 'state_vars.yml')
        with open(state_vars_file, 'w') as f:
            f.write(yaml.safe_dump(
                self.new_desired_state, default_flow_style=False))
        for play in self.playbook:
            play['tasks'].insert(0, {'include_vars': {'file': 'state_vars.yml', 'name': 'state'},
                                     'name': 'include state_vars'})

    def write_diff_vars(self):
        diff_vars_file = os.path.join(
            self.temp_dir, 'project', 'diff_vars.yml')
        with open(diff_vars_file, 'w') as f:
            f.write(yaml.safe_dump(self.state_diff, default_flow_style=False))
        for play in self.playbook:
            play['tasks'].insert(0, {'include_vars': {'file': 'diff_vars.yml', 'name': 'diff'},
                                     'name': 'include diff_vars'})

    def write_destructred_vars(self):
        for i, destructured_vars in enumerate(self.destructured_vars_list):
            diff_vars_file = os.path.join(
                self.temp_dir, 'project', f'destructured_vars_{i}.yml')
            with open(diff_vars_file, 'w') as f:
                f.write(yaml.safe_dump(
                    destructured_vars, default_flow_style=False))
        for i, play in enumerate(self.playbook):
            play['tasks'].insert(0, {'include_vars': {'file': f'destructured_vars_{i}.yml'},
                                     'name': 'include destructured_vars'})

    def write_inventory(self):
        print("inventory set to %s", self.inventory)
        with open(os.path.join(self.temp_dir, 'inventory'), 'w') as f:
            f.write(self.inventory)

    def start_ansible_playbook(self):
        # print('start_ansible_playbook')
        ansible_runner.run(private_data_dir=self.temp_dir,
                           playbook="playbook.yml",
                           quiet=True,
                           debug=True,
                           ignore_logging=True,
                           cancel_callback=self.cancel_callback,
                           finished_callback=self.finished_callback,
                           event_handler=self.runner_process_message)
        # print('finished ansible runner')
        print(self.temp_dir)

    def cancel_callback(self):
        # print('cancel_callback called')
        return self.shutdown_requested

    def finished_callback(self, runner):
        # print('finished_callback called')
        self.shutdown = True

    def runner_process_message(self, data):
        # if data.get('event', '') == 'runner_on_ok':
        # print("runner message:\n{}".format(pformat(data)))
        self.message_processor(data)
        # print(data.get('stdout', ''))

    def read_result(self):
        artifacts = glob.glob(os.path.join(
            self.temp_dir, 'artifacts', '*-*-*-*-*'))
        if len(artifacts) != 1:
            # We cannot determine the result if there is 0 or more than 1 artifact directories
            return None
        artifacts = artifacts[0]
        if not os.path.isdir(artifacts):
            # If artifacts is not a directory we cannot determine the result
            return None
        rc = os.path.join(artifacts, 'rc')
        if not os.path.exists(rc):
            # Cannot determine result if rc does not exist
            return None
        with open(rc) as f:
            return 0 == int(f.read())


def find_tasks(file_or_collection):

    if os.path.exists(file_or_collection):
        task_file = file_or_collection
    elif has_tasks(*split_collection_name(file_or_collection)):
        task_file = load_tasks(*split_collection_name(file_or_collection))
    else:
        raise Exception("No tasks found at f{file_or_collection}")
    return task_file


def desired_state_diff(monitor, secrets, project_src, current_desired_state, new_desired_state, rules, inventory, explain):
    '''
    desired_state_diff creates playbooks and runs them with ansible-runner to implement the differences
    between two version of state: current_desired_state and new_desired_state.
    '''

    # Find the difference between states

    diff = DeepDiff(current_desired_state,
                    new_desired_state, ignore_order=True)
    print(diff)

    # Find matching rules

    matching_rules = select_rules_recursive(
        diff, rules['rules'], current_desired_state, new_desired_state)
    if explain:
        print('matching_rules')
        pprint(matching_rules)

    dedup_matching_rules = deduplicate_rules(matching_rules)

    if explain:
        print('dedup_matching_rules:')
        pprint(dedup_matching_rules)

    # Build up the set of ansible-runner executions to implement the changes using the rules

    ran_rules = []

    plays = []

    destructured_vars_list = []

    for matching_rule in dedup_matching_rules:
        change_type, rule, match, value = matching_rule
        changed_subtree_path = match.groups()[0]
        action, subtree = get_rule_action_subtree(matching_rule,
                                                  current_desired_state,
                                                  new_desired_state)
        print('action', action)

        print('rule action', rule.get(ACTION_RULES[action]))

        # Experiment: Build the vars using destructuring

        destructured_vars = {}

        for name, extract_path in rule.get('vars', {}).items():
            destructured_vars[name] = extract(subtree, extract_path)

        # Experiment: Make the subtree available as node
        destructured_vars['node'] = subtree

        print('destructured_vars', destructured_vars)

        # Determine the inventory to run on

        inventory_selector = build_inventory_selector(
            rule.get('inventory_selector'))
        if inventory_selector:
            try:
                inventory_name = extract(subtree, inventory_selector)
            except KeyError:
                raise Exception(
                    f'Invalid inventory_selector {inventory_selector}')

        print('inventory_name', inventory_name)

        # Build a play using tasks or role from rule

        play = {'name': "{0} {1} {2}".format(ACTION_RULES[action], changed_subtree_path, inventory_name),
                'hosts': inventory_name,
                'gather_facts': False,
                'tasks': []}

        if 'tasks' in rule.get(ACTION_RULES[action], {}):
            play['tasks'].append({'include_tasks':
                                  {'file': find_tasks(
                                      rule.get(ACTION_RULES[action]).get('tasks'))},
                                  'name': "{0} {1}".format(ACTION_RULES[action], changed_subtree_path)})

        if 'become' in rule:
            play['become'] = rule['become']

        if explain:
            print(yaml.dump(play))
        else:

            # Run the action play

            plays.append(play)
            destructured_vars_list.append(destructured_vars)

            ran_rules.append(
                (rule, changed_subtree_path, subtree, inventory_name))

    def runner_process_message(data):
        monitor.stream.put_message(Stdout(0, now(), data.get('stdout', '')))

    PlaybookRunner(runner_process_message,
                   new_desired_state,
                   diff,
                   destructured_vars_list,
                   plays,
                   secrets,
                   project_src,
                   inventory).run()

    return ran_rules


def desired_state_discovery(monitor, secrets, project_src, current_desired_state, new_desired_state, ran_rules, inventory, explain):

    # Discovers the state of a subset of a system

    diff = DeepDiff(current_desired_state,
                    new_desired_state, ignore_order=True)

    # deep copy
    new_discovered_state = yaml.safe_load(yaml.safe_dump(new_desired_state))

    plays = []

    destructured_vars_list = []
    discovered_rules = []

    for discovery_id, (rule, changed_subtree_path, subtree, inventory_name) in enumerate(ran_rules):

        # Experiment: Build the vars using destructuring
        destructured_vars = {}

        for name, extract_path in rule.get('vars', {}).items():
            destructured_vars[name] = extract(subtree, extract_path)

        # Experiment: Make the subtree available as node
        destructured_vars['node'] = subtree
        destructured_vars['discovery_id'] = discovery_id

        print('destructured_vars', destructured_vars)

        # Build a play using tasks or role from rule

        play = {'name': f'discovery for {inventory_name} discovery_id {discovery_id}',
                'hosts': inventory_name,
                'gather_facts': False,
                'tasks': []}

        if 'tasks' in rule.get(ACTION_RULES[Action.RETRIEVE], {}):
            play['tasks'].append({'include_tasks':
                                  {'file': find_tasks(
                                      rule.get(ACTION_RULES[Action.RETRIEVE]).get('tasks'))},
                                  'name': 'include retrieve'})

            print(play)

            plays.append(play)
            destructured_vars_list.append(destructured_vars)
            discovered_rules.append(
                [discovery_id, changed_subtree_path, subtree])

    if not plays:
        return new_discovered_state

    def runner_process_message(data):
        monitor.stream.put_message(Stdout(0, now(), data.get('stdout', '')))

    runner = PlaybookRunner(runner_process_message,
                            new_desired_state,
                            diff,
                            destructured_vars_list,
                            plays,
                            secrets,
                            project_src,
                            inventory)
    result = runner.run()

    if result:

        for discovery_id, changed_subtree_path, subtree in discovered_rules:
            update_discovered_state(
                new_discovered_state, runner.temp_dir, discovery_id, changed_subtree_path, subtree)

    return new_discovered_state


def update_discovered_state(new_discovered_state, temp_dir, discovery_id, changed_subtree_path, subtree):

    discovered_state_file = os.path.join(
        temp_dir, 'project', f'discovered_state_{discovery_id}.yml')
    if os.path.exists(discovered_state_file):
        with open(discovered_state_file) as f:
            discovered_subtree_state = yaml.safe_load(f.read())
            print(changed_subtree_path)
            print(yaml.safe_dump(discovered_subtree_state, default_flow_style=False))
            print(yaml.safe_dump(subtree, default_flow_style=False))

        # List case
        match_list = re.match(r"(.*)\[(\d+)\]$", changed_subtree_path)
        if match_list:
            parent_path = match_list.groups()[0]
            index = int(match_list.groups()[1])
            extract(new_discovered_state, parent_path)[
                index] = discovered_subtree_state

        # Dict case
        match_dict = re.match(r"(.*)\['(\S+)'\]$", changed_subtree_path)
        if match_dict and not match_list:
            parent_path = match_dict.groups()[0]
            index = match_dict.groups()[1]
            extract(new_discovered_state, parent_path)[
                index] = discovered_subtree_state

        if not match_dict and not match_list:
            assert False, f"type of changed_subtree_path not supported {changed_subtree_path}"

        print(yaml.safe_dump(new_discovered_state, default_flow_style=False))


def destructure_vars(rule, subtree):

    destructured_vars = {}

    for name, extract_path in rule.get('vars', {}).items():
        destructured_vars[name] = extract(subtree, extract_path)

    return destructured_vars


def desired_state_validation(monitor, secrets, project_src, current_state, ran_rules, inventory, explain):

    plays = []

    destructured_vars_list = []
    validated_rules = []

    for rule, changed_subtree_path, subtree, inventory_name in ran_rules:

        # Experiment: Build the vars using destructuring
        destructured_vars = destructure_vars(rule, subtree)

        # Experiment: Make the subtree available as node
        destructured_vars['node'] = subtree

        print('destructured_vars', destructured_vars)

        # Build a play using tasks or role from rule

        play = {'name': f'validation for {inventory_name}',
                'hosts': inventory_name,
                'gather_facts': False,
                'tasks': []}

        if 'tasks' in rule.get(ACTION_RULES[Action.VALIDATE], {}):
            play['tasks'].append({'include_tasks':
                                  {'file': find_tasks(
                                      rule.get(ACTION_RULES[Action.VALIDATE]).get('tasks'))},
                                  'name': 'include validation'})

            print(play)

            plays.append(play)
            destructured_vars_list.append(destructured_vars)
            validated_rules.append([changed_subtree_path, subtree])

    def runner_process_message(data):
        if data.get('event', '') == 'runner_on_ok':
            event_data = data.get('event_data', {})
            if event_data.get('task_action', '') not in ['include_tasks', 'include_vars']:
                monitor.stream.put_message(ValidationTask(0, now(), event_data.get('host'),
                                                          event_data.get(
                                                              'task_action', ''),
                                                          'ok'))
        if data.get('event', '') == 'playbook_on_stats':
            event_data = data.get('event_data', {})
            for host in event_data.get('ok', {}).keys():
                monitor.stream.put_message(ValidationResult(0, now(), host, 'ok'))

        monitor.stream.put_message(Stdout(0, now(), data.get('stdout', '')))

    if not plays:
        return None

    runner = PlaybookRunner(runner_process_message,
                            current_state,
                            {},
                            destructured_vars_list,
                            plays,
                            secrets,
                            project_src,
                            inventory)
    result = runner.run()

    return result




import os
import shutil
import subprocess
import yaml
from functools import lru_cache

ANSIBLE_GALAXY = shutil.which('ansible-galaxy')


@lru_cache
def find_collection(name):
    if ANSIBLE_GALAXY is None:
        raise Exception('ansible-galaxy is not installed')
    try:
        output = subprocess.check_output(
            [ANSIBLE_GALAXY, 'collection', 'list', name], stderr=subprocess.STDOUT)
    except subprocess.CalledProcessError:
        return None
    output = output.decode()
    parts = name.split('.')
    for line in output.splitlines():
        if line.startswith('# '):
            location = line[2:]
            location = os.path.join(location, *parts)
            if os.path.exists(location):
                return location
    return None


def has_schema(collection, schema):
    if find_collection(collection) is None:
        return False
    return os.path.exists(os.path.join(find_collection(collection), 'schemas', schema) + ".yml")


def load_schema(collection, schema):
    if find_collection(collection) is None:
        return False
    location = os.path.join(find_collection(
        collection), 'schemas', schema) + ".yml"
    if not os.path.exists(location):
        raise Exception(
            f'Cannot find schema {schema} in {collection} at {location}')
    with open(location) as f:
        print(f'Loading schema from {location}')
        return yaml.safe_load(f.read())


def split_collection_name(collection_resource):
    collection, _, resource = collection_resource.rpartition(".")
    return collection, resource


def has_rules(collection, rules):
    return os.path.exists(os.path.join(find_collection(collection), 'rules', rules) + ".yml")


def load_rules(collection, rules):
    location = os.path.join(find_collection(
        collection), 'rules', rules) + ".yml"
    if not os.path.exists(location):
        raise Exception(
            f'Cannot find rules {rules} in {collection} at {location}')
    with open(location) as f:
        print(f'Loading rules from {location}')
        return yaml.safe_load(f.read())


def has_tasks(collection, tasks):
    return os.path.exists(os.path.join(find_collection(collection), 'tasks', tasks) + ".yml")


def load_tasks(collection, tasks):
    location = os.path.join(find_collection(
        collection), 'tasks', tasks) + ".yml"
    if not os.path.exists(location):
        raise Exception(
            f'Cannot find tasks {tasks} in {collection} at {location}')
    return location


def has_service(collection, service):
    return os.path.exists(os.path.join(find_collection(collection), 'services', service) + ".yml")


def load_service(collection, service):
    location = os.path.join(find_collection(collection),
                            'services', service) + ".yml"
    if not os.path.exists(location):
        raise Exception(
            f'Cannot find service {service} in {collection} at {location}')
    return location




from deepdiff import DeepDiff, extract

import re
import os

import yaml

from desired_state.util import make_matcher
from desired_state.rule import select_rules, select_rules_recursive, Action
from desired_state.diff import deduplicate_rules, get_rule_action_subtree

from pprint import pformat

HERE = os.path.abspath(os.path.dirname(__file__))


def load_rule(name):
    with open(os.path.join(HERE, 'rules', f'{name}.yml')) as f:
        return yaml.safe_load(f.read())


def load_state(name, version):
    with open(os.path.join(HERE, 'states', name, f'{version}.yml')) as f:
        return yaml.safe_load(f.read())


def run_diff_get_action(a, b, rules):

    diff = DeepDiff(a, b, ignore_order=True)
    matching_rules = select_rules_recursive(diff, rules['rules'], a, b)
    dedup_matching_rules = deduplicate_rules(matching_rules)
    assert len(dedup_matching_rules) >= 1, "No rules found"
    assert len(dedup_matching_rules) == 1, "More than on rule found: " + pformat(dedup_matching_rules)
    action, subtree = get_rule_action_subtree(dedup_matching_rules[0], a, b)
    return action, subtree


def run_diff_get_actions(a, b, rules):

    diff = DeepDiff(a, b, ignore_order=True)
    matching_rules = select_rules_recursive(diff, rules['rules'], a, b)
    dedup_matching_rules = deduplicate_rules(matching_rules)
    return [get_rule_action_subtree(x, a, b) for x in dedup_matching_rules]

def test_rule1():

    rule = yaml.safe_load(r'''
                      rule_selector: root.routers.index
                      inventory: all
                      create:
                          - tasks: create_tasks.yml
                      retrieve:
                          - tasks: get_tasks.yml
                      update:
                          - tasks: create_tasks.yml
                      delete:
                          - tasks: del_tasks.yml
                      ''')


def test_rule2():

    rule = yaml.safe_load(r'''
                      rule_selector: root.routers.index
                      inventory_selector: node.name
                      create:
                          - role: create_role
                      retrieve:
                          - role: get_role
                      update:
                          - role: update_role
                      delete:
                          - role: delete_role
                      ''')


def test_rules_change():

    t1 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R2
    ''')

    t2 = yaml.safe_load('''
    routers:
        - name: R1
        - name: R3
    ''')

    rules = yaml.safe_load(r'''
                           rules:
                            - rule_selector: root.routers.index
                              inventory_selector: node.name
                           ''')


    actions = run_diff_get_actions(t1, t2, rules)

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'name': 'R3'}


def test_rules_add():

    t1 = load_state('delete_value', 'B')
    t2 = load_state('delete_value', 'A')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'interfaces': [{'ip_address': '1.1.1.1', 'name': 'eth1'}], 'name': 'R1'}


def test_rules_delete():

    t1 = load_state('delete_value', 'A')
    t2 = load_state('delete_value', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'interfaces': [{'name': 'eth1'}], 'name': 'R1'}


def test_rules_rename():
    '''
    '''

    t1 = load_state('rename_item', 'A')
    t2 = load_state('rename_item', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'interfaces': [{'ip_address': '1.1.1.1', 'name': 'eth1'}], 'name': 'R2'}


def test_rules_list_insert_element():
    '''
    This case tests insertion into a list.  It should cause one add
    instead of multiple changes.
    '''

    t1 = load_state('add_list_value', 'B')
    t2 = load_state('add_list_value', 'A')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    print(pformat(actions))

    assert len(actions) == 1
    assert actions[0][0] == Action.CREATE
    assert actions[0][1] == {'name': 'R2'}


def test_rules_list_remove_element():
    '''
    This case tests insertion into a list.  It should cause one remove
    instead of multiple changes.
    '''

    t1 = load_state('add_list_value', 'A')
    t2 = load_state('add_list_value', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    print(pformat(actions))

    assert len(actions) == 1
    assert actions[0][0] == Action.DELETE
    assert actions[0][1] == {'name': 'R2'}


def test_rules_dictionary_add_item():
    '''
    This case tests insertion into a list.  It should cause one add
    instead of multiple changes.
    '''

    t1 = load_state('add_dict_value', 'A')
    t2 = load_state('add_dict_value', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    print(actions)

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'name': 'R1', 'router-id': '1.1.1.1'}


def test_rules_dictionary_remove_item():
    '''
    This case tests insertion into a list.  It should cause one add
    instead of multiple changes.
    '''

    t1 = load_state('add_dict_value', 'A')
    t2 = load_state('add_dict_value', 'B')
    rules = load_rule('routers_simple')

    action, subtree = run_diff_get_action(t2, t1, rules)

    assert action == Action.UPDATE
    assert subtree == {'name': 'R1'}


def test_empty_add_item():
    '''
    Tests the case to add an item to an empty file.
    '''

    t1 = load_state('empty_add_item', 'A')
    t2 = load_state('empty_add_item', 'B')
    rules = load_rule('routers_simple')

    action, subtree = run_diff_get_action(t1, t2, rules)

    assert action == Action.CREATE
    assert subtree == {'name': 'R1'}


def test_empty_remove_item():
    '''
    Tests the case to remove everything
    '''

    t1 = load_state('empty_add_item', 'B')
    t2 = load_state('empty_add_item', 'A')
    rules = load_rule('routers_simple')

    action, subtree = run_diff_get_action(t1, t2, rules)

    assert action == Action.DELETE
    assert subtree == {'name': 'R1'}


def test_reorder_list():
    '''
    '''

    t1 = load_state('reorder_list', 'A')
    t2 = load_state('reorder_list', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    assert len(actions) == 0


def test_rules_rename_key():
    '''
    Changing a key in a subtree should cause a single update.
    '''

    t1 = load_state('rename_key', 'A')
    t2 = load_state('rename_key', 'B')
    rules = load_rule('routers_simple')

    actions = run_diff_get_actions(t1, t2, rules)

    print(pformat(actions))

    assert len(actions) == 1
    assert actions[0][0] == Action.UPDATE
    assert actions[0][1] == {'interfaces': [{'ip_address': '1.1.1.1', 'name': 'eth1'}], 'label': 'R1'}


def test_rules_rename_key2():
    '''
    Changing the root key of a subtree should cause a delete and a create operation.
    '''

    t1 = load_state('rename_key2', 'A')
    t2 = load_state('rename_key2', 'B')
    rules = load_rule('router_switch')

    actions = run_diff_get_actions(t1, t2, rules)

    print(pformat(actions))

    assert len(actions) == 2
    assert actions[0][0] == Action.CREATE
    assert actions[0][1] == {'interfaces': [{'ip_address': '1.1.1.1', 'name': 'eth1'}], 'name': 'R1'}
    assert actions[1][0] == Action.DELETE
    assert actions[1][1] == {'interfaces': [{'ip_address': '1.1.1.1', 'name': 'eth1'}], 'name': 'R1'}
